/* Asterisk Manager Proxy
   2005-06-09
   Copyright (c) 2005 David C. Troy <dave@popvox.com>

   This program is free software, distributed under the terms of
   the GNU General Public License.

   Thanks to Bryan Ballard <ballard@netsolus.com> for XML fixes.

   TODO:  Add support for multiple Asterisk servers
	  Add support for XML input
	  Add support for Input/Output Format Abstraction
	  Figure out a more secure proxy-level authentication model
	  Add support for Extensible Proxy Command language, Proxy Macros 
	  Add idea of loglevels to logmsg/debugmsg; kill debugmsg
	  Cache I/O on connections that die
*/

#include "astmanproxy.h"

#define PROXY_BANNER	"Asterisk Call Manager Proxy"
#define PROXY_SHUTDOWN	"Proxy Shutdown"
#define DATEFORMAT	"%b %e %T"

extern void ReadConfig( void );
extern FILE *OpenLogfile( void );
extern int SetProcUID( void );

int WriteClient(struct mansession *, struct message *);
int WriteXMLClient(struct mansession *s, struct message *m);
int get_input(struct mansession *, char *);
int ConnectAsterisk();

struct mansession *sessions;
struct mansession ast;
struct proxyconfig pc;

pthread_mutex_t sessionlock;
pthread_mutex_t loglock;
pthread_mutex_t debuglock;
static int asock = -1;
FILE *proxylog;
int debug = 0;

void debugmsg (const char *fmt, ...)
{
    va_list ap;

    time_t t;
    struct tm tm;
    char date[80];

    time(&t);
    localtime_r(&t, &tm);
    strftime(date, sizeof(date), DATEFORMAT, &tm);

    pthread_mutex_lock(&debuglock);
    va_start(ap, fmt);
    printf("%s: ", date);
    vprintf(fmt, ap);
    printf("\n");
    va_end(ap);
    pthread_mutex_unlock(&debuglock);
}


void logmsg (const char *fmt, ...)
{
    va_list ap;

    time_t t;
    struct tm tm;
    char date[80];

    time(&t);
    localtime_r(&t, &tm);
    strftime(date, sizeof(date), DATEFORMAT, &tm);

    if (proxylog) {
        pthread_mutex_lock(&loglock);
        va_start(ap, fmt);
        fprintf(proxylog, "%s: ", date);
        vfprintf(proxylog, fmt, ap);
        fprintf(proxylog, "\n");
        va_end(ap);
        fflush(proxylog);
        pthread_mutex_unlock(&loglock);
    }
}

/* Recursive thread safe replacement of inet_ntoa */
const char *ast_inet_ntoa(char *buf, int bufsiz, struct in_addr ia)
{
        return inet_ntop(AF_INET, &ia, buf, bufsiz);
}

void hup(int sig) {
    if (proxylog) {
       fflush(proxylog);
       fclose(proxylog);
    }
    proxylog = OpenLogfile();
    logmsg("Received HUP -- reopened log");
}

void leave(int sig) {
    struct mansession *c;
    struct message m;

    memset(&m, 0, sizeof(&m));
    strcpy(m.headers[0], PROXY_SHUTDOWN);
    m.source = SRC_PROXY;
    m.hdrcount++;

    if (debug)
        debugmsg("Shutting Down Asterisk Connection");
    close(ast.fd);
    
    if (debug)
        debugmsg("Notifying and closing clients");
    pthread_mutex_lock (&sessionlock);
    while (sessions) {
        c = sessions;
        sessions = sessions->next;
        WriteClient(c, &m);
        if (debug)
           debugmsg("Closing client.");
        close(c->fd);
        pthread_mutex_destroy(&c->lock);
        free(c);
    }
    pthread_mutex_unlock (&sessionlock);

    if (debug)
        debugmsg("Closing Listener Socket");
    close(asock);

    if(debug)
        debugmsg("Done!\n");
    logmsg("Proxy stopped; shutting down.");

    fclose(proxylog);
    pthread_mutex_destroy(&sessionlock);
    pthread_mutex_destroy(&loglock);
    pthread_mutex_destroy(&debuglock);
    /*pthread_exit(NULL);*/
    exit(sig);
}

char *astman_get_header(struct message *m, char *var)
{
        char cmp[80];
        int x;
        snprintf(cmp, sizeof(cmp), "%s: ", var);
        for (x=0;x<m->hdrcount;x++)
                if (!strncasecmp(cmp, m->headers[x], strlen(cmp)))
                        return m->headers[x] + strlen(cmp);
        return "";
}

void Version( void )
{
    printf("astmanproxy: Version %s, (C) David C. Troy\n", PROXY_VERSION);
    return;
}

void Usage( void )
{
    printf("Usage: astmanproxy [-d|-h|-v]\n");
    printf(" -d : Start in Debug Mode\n");
    printf(" -h : Displays this message\n");
    printf(" -v : Displays version information\n");
    printf("Start with no options to run as daemon\n");
    return;
}

void *ProxyListClients(struct mansession *s) {
    struct message m;
    struct mansession *c;
    char iabuf[INET_ADDRSTRLEN];

    memset(&m, 0, sizeof(&m));
    m.source = SRC_PROXY;
    strcpy(m.headers[0], "ProxyResponse: Success");
    m.hdrcount++;

    c = sessions;
    while (c && (m.hdrcount < MAX_HEADERS - 1) ) {
        sprintf(m.headers[m.hdrcount], "ProxyClient: %s (%s)", ast_inet_ntoa(iabuf, sizeof(iabuf), c->sin.sin_addr), c->actionid);
        m.hdrcount++;
        c = c->next;
    }
    WriteClient(s, &m);
    return 0;
}

void *ProxySetOutputFormat(struct mansession *s, struct message *m) {
    struct message mo;
    char *value;
    int i;

    value = astman_get_header(m, "OutputFormat");
    if ( !strcasecmp(value, "XML") )
        i = 1;
    else if ( !strcasecmp(value, "Standard") )
        i = 0;
    else
        return 0;
        /* TODO: Return error from proxy for invalid input */

    pthread_mutex_lock(&s->lock);
    s->outputformat = i;
    pthread_mutex_unlock(&s->lock);

    memset(&mo, 0, sizeof(&mo));
    mo.source = SRC_PROXY;
    strcpy(mo.headers[0], "ProxyResponse: Success");
    mo.hdrcount++;
    sprintf(mo.headers[mo.hdrcount], "OutputFormat: %s", s->outputformat ? "XML" : "Standard");
    mo.hdrcount++;

    WriteClient(s, &mo);

    return 0;
}

void *ProxySetAutoFilter(struct mansession *s, struct message *m) {
    struct message mo;
    char *value;
    int i;

    value = astman_get_header(m, "AutoFilter");
    if ( !strcasecmp(value, "on") )
        i = 1;
    else
        i = 0;
    pthread_mutex_lock(&s->lock);
    s->autofilter = i;
    pthread_mutex_unlock(&s->lock);

    memset(&mo, 0, sizeof(&mo));
    mo.source = SRC_PROXY;
    strcpy(mo.headers[0], "ProxyResponse: Success");
    mo.hdrcount++;
    sprintf(mo.headers[mo.hdrcount], "AutoFilter: %d", s->autofilter);
    mo.hdrcount++;

    WriteClient(s, &mo);

    return 0;
}

static void destroy_session(struct mansession *s)
{
    struct mansession *cur, *prev = NULL;
    char iabuf[INET_ADDRSTRLEN];

        pthread_mutex_lock(&sessionlock);
        cur = sessions;
        while(cur) {
                if (cur == s)
                        break;
                prev = cur;
                cur = cur->next;
        }
        if (cur) {
                if (prev)
                        prev->next = cur->next;
                else
                        sessions = cur->next;
                logmsg("Connection closed: %s", ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr));
                if (s->fd > -1)
                        close(s->fd);
                pthread_mutex_destroy(&s->lock);
                free(s);
        } else
                logmsg("Trying to delete non-existent session %p?\n", s);
        pthread_mutex_unlock(&sessionlock);

}

void *ProxyLogin(struct mansession *s) {
    struct message m;

    /* Send back dummy output for clients that insist on authenticating the old way */
    /* Mostly, we want to avoid sending this to Asterisk */

    /* Response: Success			*/
    /* Message: Authentication accepted		*/

    memset(&m, 0, sizeof(&m));
    m.source = SRC_PROXY;
    strcpy(m.headers[0], "Response: Success");
    m.hdrcount++;
    strcpy(m.headers[1], "Message: Authentication Accepted");
    m.hdrcount++;

    WriteClient(s, &m);

    return NULL;
}

void *ProxyLogoff(struct mansession *s) {
    struct message m;

    memset(&m, 0, sizeof(&m));
    m.source = SRC_PROXY;
    strcpy(m.headers[0], "Goodbye: Y'all come back now, y'hear?");
    m.hdrcount++;

    WriteClient(s, &m);

    destroy_session(s);
    if (debug)
        debugmsg("Client logged off - exiting thread");
    pthread_exit(NULL);
    return NULL;
}

int WriteSTDClient(struct mansession *s, struct message *m) {
    int i;
    char outstring[MAX_LEN];

    for (i=0; i<m->hdrcount; i++) {
        strcpy(outstring, m->headers[i]);
        strcat(outstring, "\r\n");
        pthread_mutex_lock(&s->lock);
        write(s->fd, outstring, strlen(outstring) );
        pthread_mutex_unlock(&s->lock);
    }
    pthread_mutex_lock(&s->lock);
    write(s->fd, "\r\n", 2);
    pthread_mutex_unlock(&s->lock);

    return 0;
}

int WriteClient(struct mansession *s, struct message *m) {
    if ( s->outputformat == 1)
        WriteXMLClient(s, m);
    else
        WriteSTDClient(s, m);
    return 0;
}

int WriteClients(struct message *m) {
    struct mansession *c;
    char *actionid;

    c = sessions;
    while (c) {
        if (c->autofilter && c->actionid) {
            actionid = astman_get_header(m, "ActionID");
            if ( !strcmp(actionid, c->actionid) )
                WriteClient(c, m);
        } else
            WriteClient(c, m);
        c = c->next;
    }
    return 1;
}

int WriteAsterisk(struct message *m) {
    int i;
    char outstring[MAX_LEN];

    /* Check for empty block -- Don't pester Asterisk if it is one*/
    if (!m->hdrcount && !m->headers[0][0])
	return 1;

    for (i=0; i<m->hdrcount; i++) {
        sprintf(outstring, "%s\r\n", m->headers[i]);
        pthread_mutex_lock(&ast.lock);
        write(ast.fd, outstring, strlen(outstring) );
        pthread_mutex_unlock(&ast.lock);
    }
    pthread_mutex_lock(&ast.lock);
    write(ast.fd, "\r\n", 2);
    pthread_mutex_unlock(&ast.lock);
    return 1;
}

void *proxyerror_do(struct message *m, struct mansession *s)
{
    struct message mo;

    memset(&mo, 0, sizeof(&mo));
    mo.source = SRC_PROXY;
    strcpy(mo.headers[0], "ProxyResponse: Error");
    mo.hdrcount++;
    strcpy(mo.headers[mo.hdrcount], "Message: Invalid Proxy Action");
    mo.hdrcount++;

    WriteClient(s, &mo);

    return 0;
}

void *proxyaction_do(char *proxyaction, struct message *m, struct mansession *s)
{
    if (!strcasecmp(proxyaction,"SetOutputFormat")) {
        ProxySetOutputFormat(s, m);
    } else if (!strcasecmp(proxyaction,"SetAutoFilter")) {
        ProxySetAutoFilter(s, m);
    } else if (!strcasecmp(proxyaction,"ListClients"))
        ProxyListClients(s);
    else if (!strcasecmp(proxyaction,"Logoff"))
        ProxyLogoff(s);
    else
	proxyerror_do(m, s);

    return 0;
}

void *setactionid(char *actionid, struct message *m, struct mansession *s)
{
    pthread_mutex_lock(&s->lock);
    strncpy(s->actionid, actionid, MAX_LEN);
    pthread_mutex_unlock(&s->lock);

    return 0;
}

void *session_do(struct mansession *s)
{
    struct message m;
    int res;
    char *proxyaction, *actionid, *action, *key;

    memset(&m, 0, sizeof(&m));
    sprintf(m.headers[0], "%s/%s", PROXY_BANNER, PROXY_VERSION);
    m.source = SRC_PROXY;
    m.hdrcount++;
    WriteClient(s, &m);

    memset(&m, 0, sizeof(&m));
    m.source = SRC_CLIENT;
    for (;;) {
        res = get_input(s, m.headers[m.hdrcount]);
        if (res > 0) {
            /* Strip trailing \r\n
            if (strlen(m.headers[m.hdrcount]) < 2)
                continue;
            m.headers[m.hdrcount][strlen(m.headers[m.hdrcount]) - 2] = '\0'; */
            if (*(m.headers[m.hdrcount]) == '\0' ) {
		/* Check for anything that requires proxy-side processing */
                if (pc.key && !s->authenticated) {
                    key = astman_get_header(&m, "ProxyKey");
                    if (!strcmp(key, pc.key) ) {
                       pthread_mutex_lock(&s->lock);
                       s->authenticated = 1;
                       pthread_mutex_unlock(&s->lock);
                    } else
                        break;
                }
                proxyaction = astman_get_header(&m, "ProxyAction");
                actionid = astman_get_header(&m, "ActionID");
                action = astman_get_header(&m, "Action");
                if ( !strcasecmp(action, "Login") )
                    ProxyLogin(s);
                else if ( !strcasecmp(action, "Logoff") )
                    ProxyLogoff(s);
                else if ( !(*proxyaction == '\0') )
                    proxyaction_do(proxyaction, &m, s);
                else {
                    if ( !(*actionid == '\0') )
                        setactionid(actionid, &m, s);
                    if ( !WriteAsterisk(&m) )
                        break;
                }
                memset(&m, 0, sizeof(&m));
            } else if (m.hdrcount < MAX_HEADERS - 1)
                m.hdrcount++;
        } else if (res < 0)
            break;
   }

   destroy_session(s);
   if (debug)
       debugmsg("Exiting session_do thread");
   pthread_exit(NULL);
   return NULL;
}

void *HandleAsterisk(void *nullptr)
{
    struct message m;
    int res;

    memset(&m, 0, sizeof(&m));
    m.source = SRC_ASTERISK;
    for (;;) {
        res = get_input(&ast, m.headers[m.hdrcount]);
        if (res > 0) {
            /* Strip trailing \r\n
            if (strlen(m.headers[m.hdrcount]) < 2)
                continue;
            m.headers[m.hdrcount][strlen(m.headers[m.hdrcount]) - 2] = '\0'; */
            if (debug)
                debugmsg("Received from Asterisk: %s", m.headers[m.hdrcount]);
            if (*(m.headers[m.hdrcount]) == '\0' ) {
                if (!WriteClients(&m))
                    break;
                memset(&m, 0, sizeof(&m));
            } else if (m.hdrcount < MAX_HEADERS - 1)
                m.hdrcount++;
        } else if (res < 0) {
            if (debug)
                debugmsg("Lost Asterisk Connection; reconnecting");
                if ( !ConnectAsterisk() )
                    leave(1);
        }
   }

   pthread_exit(NULL);
   return NULL;
}

int ConnectAsterisk() {
    char buf[BUFSIZE];

    if (debug)
        debugmsg("Connecting to Asterisk Manager");

    bzero((char *) &ast.sin,sizeof(ast.sin)); 
    ast.sin.sin_family = AF_INET;
    ast.sin.sin_addr.s_addr = inet_addr(pc.ast_addr); 
    ast.sin.sin_port = htons(pc.ast_port);
    if((ast.fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) return(0);

    for ( ;; ) {
        if(connect(ast.fd, (struct sockaddr *) &ast.sin, sizeof(ast.sin)) < 0) {
            if (debug)
                debugmsg("Cannot connect to Asterisk Manager!  Retrying.");
            sleep(1);
        } else { 
            sprintf(buf, "Action: Login\r\nUsername: %s\r\nSecret: %s\r\nEvents: %s\r\n\r\n", pc.ast_user, pc.ast_pass, 
                pc.events ? "on" : "off" );
            write(ast.fd, buf, strlen(buf));
            if (debug)
                debugmsg("Connected to asterisk; logging in");
            break;
        }
    }
    return 1;
}

int get_input(struct mansession *s, char *output)
{
        /* output must have at least sizeof(s->inbuf) space */
        int res;
        int x;
        struct pollfd fds[1];
        char iabuf[INET_ADDRSTRLEN];

        /* Look for \r\n from the front, our preferred end of line */
        for (x=1;x<s->inlen;x++) {
                if ((s->inbuf[x] == '\n') && (s->inbuf[x-1] == '\r')) {
                        /* Copy output data up to and including \r\n */
                        memcpy(output, s->inbuf, x - 1);
                        /* Add trailing \0 */
                        output[x-1] = '\0';
                        /* Move remaining data back to the front */
                        memmove(s->inbuf, s->inbuf + x + 1, s->inlen - x);
                        s->inlen -= (x + 1);
                        return 1;
                }
        }

        /* Then look for \n */
        for (x=1;x<s->inlen;x++) {
                if ( (s->inbuf[x] == '\n') ) {
                        memcpy(output, s->inbuf, x);
                        /* Add trailing \0 */
                        output[x] = '\0';
                        memmove(s->inbuf, s->inbuf + x + 1, s->inlen - x);
                        s->inlen -= (x + 1);
                        return 1;
                }
        }


        if (s->inlen >= sizeof(s->inbuf) - 1) {
            if (debug)
                debugmsg("Warning: Got long line with no end from %s: %s\n", ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr), s->inbuf);
            s->inlen = 0;
        }
        fds[0].fd = s->fd;
        fds[0].events = POLLIN;
        res = poll(fds, 1, -1);
        if (res < 0 && debug) {
                debugmsg("Select returned error");
        } else if (res > 0) {
                pthread_mutex_lock(&s->lock);
                res = read(s->fd, s->inbuf + s->inlen, sizeof(s->inbuf) - 1 - s->inlen);
                pthread_mutex_unlock(&s->lock);
                if (res < 1)
                        return -1;
        }
        s->inlen += res;
        s->inbuf[s->inlen] = '\0';
        return 0;
}

static void *accept_thread()
{
        int as;
        struct sockaddr_in sin;
        int sinlen;
        struct mansession *s;
        struct protoent *p;
        int arg = 1;
        int flags;
        pthread_attr_t attr;
        char iabuf[INET_ADDRSTRLEN];

        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

        for (;;) {
                sinlen = sizeof(sin);
                as = accept(asock, (struct sockaddr *)&sin, &sinlen);
                if (as < 0) {
                        logmsg("Accept returned -1: %s\n", strerror(errno));
                        continue;
                }
                p = (struct protoent *)getprotobyname("tcp");
                if( p ) {
                        if( setsockopt(as, p->p_proto, TCP_NODELAY, (char *)&arg, sizeof(arg) ) < 0 ) {
                                logmsg("Failed to set listener tcp connection to TCP_NODELAY mode: %s\n", strerror(errno));
                        }
                }
                s = malloc(sizeof(struct mansession));
                if ( !s ) {
                        logmsg("Failed to allocate listener session: %s\n", strerror(errno));
                        continue;
                }
                memset(s, 0, sizeof(struct mansession));
                memcpy(&s->sin, &sin, sizeof(sin));

                /* For safety, make sure socket is non-blocking */
                flags = fcntl(as, F_GETFL);
                fcntl(as, F_SETFL, flags | O_NONBLOCK);

                pthread_mutex_init(&s->lock, NULL);
                s->fd = as;
                s->outputformat = pc.outputformat;
                s->autofilter = pc.autofilter;
                pthread_mutex_lock(&sessionlock);
                s->next = sessions;
                sessions = s;
                pthread_mutex_unlock(&sessionlock);

                logmsg("Connection received from %s", ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr));
                if (debug)
                   debugmsg("Connection received from %s", ast_inet_ntoa(iabuf, sizeof(iabuf), s->sin.sin_addr));

                if (pthread_create(&s->t, &attr, (void *)session_do, s))
                        destroy_session(s);
        }
        pthread_attr_destroy(&attr);
        return NULL;
}

int main(int argc, char *argv[])
{
    struct sockaddr_in serv_sock_addr, client_sock_addr; 
    int cli_addrlen;
    pthread_t ast_thread;
    struct linger lingerstruct;        /* for socket reuse */
    int flag;                          /* for socket reuse */
    pid_t pid;
    char i;

    /* Figure out if we are in debug mode, handle other switches */
    while (( i = getopt( argc, argv, "dhv" ) ) != EOF )
    {
        switch( i ) {
            case 'd':
                debug = 1;
                break;
            case 'h':
                Usage();
                exit(0);
            case 'v':
                Version();
                exit(0);
            case '?':
                Usage();
                exit(1);
        }
    }

    ReadConfig();
    proxylog = OpenLogfile();

    if (SetProcUID()) {
        fprintf(stderr,"Cannot set user/group!  Check proc_user and proc_group config setting!\n");
        exit(1);
    }

    /* Initialize global client list */
    sessions = NULL;

    /* If we are not in debug mode, then fork to background */
    if (!debug) {
        if ( (pid = fork()) < 0)
            exit( 1 );
        else if ( pid > 0)
            exit( 0 );
    }

    /* Setup signal handlers */
    (void) signal(SIGINT,leave);
    (void) signal(SIGHUP,hup);
    (void) signal(SIGTERM,leave);

    /* Connect to Asterisk */
    ConnectAsterisk();

    /* Setup listener socket... */
    if ((asock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        fprintf(stderr,"Cannot create listener socket!\n");
        exit(1);
    }
    bzero((char *) &serv_sock_addr, sizeof serv_sock_addr );
    serv_sock_addr.sin_family = AF_INET;

    if ( !strcmp(pc.listen_addr,"*") )
        serv_sock_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    else
        serv_sock_addr.sin_addr.s_addr = inet_addr( pc.listen_addr);
    serv_sock_addr.sin_port = htons((short)pc.listen_port);

    /* Set listener socket re-use options */
    setsockopt(asock, SOL_SOCKET, SO_REUSEADDR, (void *)&flag, sizeof(flag));
    lingerstruct.l_onoff = 1;
    lingerstruct.l_linger = 5;
    setsockopt(asock, SOL_SOCKET, SO_LINGER, (void *)&lingerstruct, sizeof(lingerstruct));
    
    if (bind(asock, (struct sockaddr *) &serv_sock_addr, sizeof serv_sock_addr ) < 0) {
        fprintf(stderr,"Cannot bind to listener socket!\n");
        exit(1);
    }

    /* Initialize global mutexes */
    pthread_mutex_init(&sessionlock, NULL);
    pthread_mutex_init(&loglock, NULL);
    pthread_mutex_init(&debuglock, NULL);

    /* Launch Asterisk handling thread... */
    if (debug)
        debugmsg("Launching Asterisk Handler");
    if (pthread_create( &ast_thread, NULL, (void *)HandleAsterisk, NULL))
        exit(1);

    listen(asock, 5);
    cli_addrlen = sizeof(client_sock_addr);
    if (debug)
        debugmsg("Listening for connections");
    logmsg("Proxy Started: Listening for connections");

    /* Launch listener thread */
    accept_thread();

    pthread_exit(NULL);
    exit(0);
}

